/*
 * This example captures a set of frames from SUNWrtvc and saves the captured
 * XIL compressed image seqence (CIS) into a file.
 */

#include <stdio.h>
#include <stdlib.h>
#include <xil/xil.h>

/* - - - - - - - - - - - - - - - - - - - - - - */

#define GETOPTARGS "o:C:s:I:f:i:D:E"

void
usage(char *name, char *badparam)
{
    if (badparam)
	fprintf(stderr, "ERROR: bad parameter [%s]\n", badparam);

    fprintf(stderr, "\
Usage: %s [options]\n\n\
  Legal Movie Options: \n\
        -o %%s  - save bitstream in file         NULL \n\
        -C %%s  - Compression type               CellB\n\
                    [ CellB | Mpeg1 | Jpeg ]\n\
  Legal Options:                               Default\n\
        -s %%d  - shrink factor                  2\n\
        -I %%d  - select input channel           1\n\
        -f %%d  - frames to capture              100\n\
        -i %%d  - frames to skip between capture 0\n\
        -D %%s  - name of rtvc device            /dev/rtvc0\n\
  Legal Switches:\n\
        -E      - exit on first error\n\
", name);

    exit(1);
}

/* - - - - - - - - - - - - - - - - - - - - - - */
/* Error handler installed iff -E specified */
Xil_boolean
error_handler(XilError error)
{
    xil_call_next_error_handler(error);

    fprintf(stderr, "\n***ERROR received; example exiting\n");
    exit(1);
}


/* - - - - - - - - - - - - - - - - - - - - - - */
void
write_frames(FILE * df, XilCis cis)
{
    Xil_unsigned8 *data;

    int nframes, nbytes;

    while (xil_cis_has_frame(cis)) {
	data = (void *) xil_cis_get_bits_ptr(cis, &nbytes, &nframes);

	if (df)
	    fwrite(data, 1, nbytes, df);
    }
}

/* - - - - - - - - - - - - - - - - - - - - - - */

main(int argc, char **argv)
{

    /* XIL variables */
    XilSystemState state;

    XilImage rtvc_image;

    XilImage scaled_image;

    XilDataType datatype;

    Xil_boolean error_out = FALSE;

    XilCis cis;

    XilMpeg1Pattern mpattern;

    char pattern[32];

    unsigned int width, height, nbands;

    int frames;

    int total_frames = 100;

    int image_skip = 0;

    int doscale = 2;

    int i;

    double t;

    struct timeval t0, t1;

    int port = 1;

    int c;

    char *movie_file = NULL;

    char *devname = NULL;

    FILE *df;

    enum compressor_type {
	JPEG, MPEG, CELLB
    };

    enum compressor_type do_cis = CELLB;

    char *cis_type = "CellB";

    strcpy(pattern, "IIII");

    while ((c = getopt(argc, argv, GETOPTARGS)) != EOF)
	switch (c) {
	case 's':
	    doscale = atoi(optarg);
	    break;
	case 'I':
	    port = atoi(optarg);
	    break;
	case 'D':
	    devname = optarg;
	    break;
	case 'i':
	    image_skip = atoi(optarg);
	    break;
	case 'f':
	    total_frames = atoi(optarg);
	    break;
	case 'C':
	    if (!strcasecmp(optarg, "Mpeg1")) {
		do_cis = MPEG;
		cis_type = "Mpeg1";
	    }
	    else if (!strcasecmp(optarg, "CellB")) {
		do_cis = CELLB;
		cis_type = "CellB";
	    }
	    else if (!strcasecmp(optarg, "Jpeg")) {
		do_cis = JPEG;
		cis_type = "Jpeg";
	    }
	    else
		usage(argv[0], optarg);
	    break;
	case 'o':
	    movie_file = optarg;
	    break;
	case 'E':
	    error_out = TRUE;
	    break;
	case '?':
	default:
	    usage(argv[0], NULL);
	}

    /* expect no args left over */
    if (argv[optind]) {
	usage(argv[0], argv[optind]);
    }

    if (!movie_file) {
	fprintf(stderr, "\n***ERROR: must set `-o' option for output movie filename.\n\n");
	usage(argv[0], NULL);
    }

    df = fopen(movie_file, "w");
    if (df == NULL) {
	fprintf(stderr, "failed to open output movie file.\n");
	exit(1);
    }

    state = xil_open();
    if (state == NULL) {
	fprintf(stderr, "unable to open xil library\n");
	exit(1);
    }

    if (error_out)
	xil_install_error_handler(state, error_handler);

    /* XIL_API_MAJOR_VERSION is #define'd beginning with XIL release 1.2 */

#ifdef XIL_API_MAJOR_VERSION
    {
	XilDevice device;

	if (!(device = xil_device_create(state, "SUNWrtvc"))) {
	    fprintf(stderr, "Unable to create a device object\n");
	    xil_close(state);
	    exit(1);
	}

	xil_device_set_value(device, "DEVICE_NAME", (void *) devname);
	rtvc_image = xil_create_from_device(state, "SUNWrtvc", device);
	xil_device_destroy(device);
    }
#else
    rtvc_image = xil_create_from_device(state, "SUNWrtvc", (void *) devname);
#endif

    if (rtvc_image == NULL) {
	fprintf(stderr, "failed to open SUNWrtvc device\n");
	xil_close(state);
	exit(1);
    }

    xil_set_device_attribute(rtvc_image, "PORT_V", (void *) port);

    {
	int format;

	xil_get_device_attribute(rtvc_image, "FORMAT_V", (void **) &format);
	if (format == 0) {
	    fprintf(stderr, "Unknown video format, exiting.\n");
	    xil_close(state);
	    exit(1);
	}
    }

    if (image_skip)
	xil_set_device_attribute(rtvc_image, "IMAGE_SKIP",
				 (void *) image_skip);

    xil_get_info(rtvc_image, &width, &height, &nbands, &datatype);

    if (doscale) {
	width = (int) (((float) width / doscale) + 0.5);
	height = (int) (((float) height / doscale) + 0.5);
	scaled_image = xil_create(state, width, height, nbands, XIL_BYTE);
    }

    /* Create a cis to put the movie frames in. */
    cis = xil_cis_create(state, cis_type);

    if (do_cis == MPEG) {
	if (doscale != 2) {
	    fprintf(stderr, "ERROR: scalefactor is: %d, should be 2 for Mpeg1\n",
		    doscale);
	    xil_cis_destroy(cis);
	    xil_close(state);
	    exit(1);
	}

	/* Mpeg1 currently needs to keep at least 6 frames around */
	xil_cis_set_keep_frames(cis, 6);
	xil_cis_set_max_frames(cis, 6);

	/* set the Mpeg1 compression pattern */
	mpattern.pattern = pattern;
	mpattern.repeat_count = 0;

	xil_cis_set_attribute(cis, "COMPRESSOR_PATTERN", (void *) &mpattern);

	/*
	 * SunVideo's Mpeg1 Variable Bit Rate Compression quality is set by
	 * the device attribute MPEG_FIXED_QUANT. Valid values range from 1
	 * to 31 where 1 yields the highest quality (and the largest bit
	 * rate).
	 */

	xil_set_device_attribute(rtvc_image, "MPEG_FIXED_QUANT", (void *) (1));

	xil_cis_set_attribute(cis, "COMPRESSOR_INSERT_VIDEO_SEQUENCE_END", (void *) TRUE);
    }
    else {

	/*
	 * set the buffers to make sure that you don't keep information you
	 * do not need.  Mpeg1 needs more frames.
	 */
	xil_cis_set_keep_frames(cis, 1);
	xil_cis_set_max_frames(cis, 1);
    }

    /*
     * If JPEG, set an attribute needed for fast decompression of the
     * resulting movie.
     */
    if (do_cis == JPEG)
	xil_cis_set_attribute(cis, "ENCODE_411_INTERLEAVED", (void *) TRUE);

    for (frames = 0; frames < total_frames; frames++) {
	if (doscale) {
	    xil_scale(rtvc_image, scaled_image, "nearest",
		      1.0 / (float) doscale, 1.0 / (float) doscale);
	}
	else {
	    scaled_image = rtvc_image;
	}

	xil_compress(scaled_image, cis);
	xil_cis_sync(cis);
	xil_toss(scaled_image);

	write_frames(df, cis);

	/* don't include overhead of first frame in frame rate calculation */
	if (frames == 0) {
	    gettimeofday(&t0, NULL);
	}
    }

    /* done, empty buffer if any frames remain */
    xil_cis_flush(cis);
    write_frames(df, cis);

    /* XIL_API_MAJOR_VERSION is #define'd beginning with XIL release 1.2 */

#ifndef XIL_API_MAJOR_VERSION
    /* XIL versions prior to 1.2 do not correctly implement */
    /* MPEG COMPRESSOR_INSERT_VIDEO_SEQUENCE_END */
    if (do_cis == MPEG && df) {
	unsigned int sec = 0x000001B7;

	fwrite((char *) &sec, 1, 4, df);
    }
#endif

    /* calculate frame rate */
    gettimeofday(&t1, NULL);
    t = (t1.tv_sec - t0.tv_sec) + (t1.tv_usec - t0.tv_usec) * 1e-6;
    printf("%d frames in %g seconds, %g frames/second\n",
	   frames - 1, t, (frames - 1) / t);

    xil_close(state);

    return 0;
}
