/*
 * This example reads an image from SUNWrtvc and displays the image in a
 * Xwindow
 */

#include <stdio.h>
#include <stdlib.h>
#include <xil/xil.h>

#include "ddither.h"

Atom closedownAtom;

/* - - - - - - - - - - - - - - - - - - - - - - */

#define GETOPTARGS "s:I:f:w:h:i:m:x:y:D:gEd"

void 
usage(char *name, char *badparam)
{
    if (badparam)
	fprintf(stderr, "ERROR: bad parameter [%s]\n", badparam);

    fprintf(stderr, "\
Usage: %s [options]\n\n\
Legal Options:					 Default\n\
	-s %%d	- shrink factor			 2\n\
	-I %%d	- select input channel		 1\n\
	-f %%d	- frames to capture		 100\n\
	-w %%d	- capture subimage width	 640\n\
	-h %%d	- capture subimage height	 480\n\
	-i %%d	- frames to skip between capture 0\n\
        -m %%d	- max frames to buffer		 2\n\
        -x %%d	- x window location		 (random)\n\
        -y %%d	- y window location		 (random)\n\
	-D %%s	- name of rtvc device		 /dev/rtvc0\n\
Legal Switches:\n\
	-g	- enable grayscale mode \n\
        -E      - exit on first error  \n\
	-d	- don't display result of capture \n\
", name);

    exit(1);
}

/* - - - - - - - - - - - - - - - - - - - - - - */

/*
 * Error handler installed iff -E specified
 */
Xil_boolean 
error_handler(XilError error)
{
    xil_call_next_error_handler(error);
    fprintf(stderr, "\n***ERROR received; example exiting\n");
    exit(1);
}

main(int argc, char **argv)
{
    XilSystemState state;

    XilImage rtvc_image, rtvc_luma, display;

    XilImage rescaled_image, scaled_image, single_band;

    unsigned int width, height, nbands;

    int child_width = 0;

    int child_height = 0;

    int winx = 0;

    int winy = 0;

    int xyflag = 0;

    XilDataType datatype;

    Display *xdisplay;

    Window xwindow;

    XEvent event;

    int display_depth;

    XilLookup yuv_to_rgb;

    XilLookup colorcube;

    XilLookup graylookup;

    XilDitherMask mask;

    XilColorspace cspace_ycc601;

    XilColorspace cspace_rgb709;

    int frames;

    int total_frames = 100;

    int image_skip = 0;

    int max_buffers = -1;

    int doscale = 2;

    float scale_factor;

    int grayscale = 0;

    float rescale_values[3];

    float offset_values[3];

    double t;

    struct timeval t0, t1;

    int value;

    char *type;

    int port = 1;

    Xil_boolean nodisplay = FALSE;

    Xil_boolean error_out = FALSE;

    char *devname = "/dev/rtvc0";

    int c;


    while ((c = getopt(argc, argv, GETOPTARGS)) != EOF)
	switch (c) {
	case 's':
	    doscale = atoi(optarg);
	    break;
	case 'I':
	    port = atoi(optarg);
	    break;
	case 'D':
	    devname = optarg;
	    break;
	case 'i':
	    image_skip = atoi(optarg);
	    break;
	case 'm':
	    max_buffers = atoi(optarg);
	    break;
	case 'x':
	    winx = atoi(optarg);
	    xyflag = 1;
	    break;
	case 'y':
	    winy = atoi(optarg);
	    xyflag = 1;
	    break;
	case 'f':
	    total_frames = atoi(optarg);
	    break;
	case 'w':
	    child_width = atoi(optarg);
	    break;
	case 'h':
	    child_height = atoi(optarg);
	    break;
	case 'g':
	    grayscale = 1;
	    break;
	case 'd':
	    nodisplay = TRUE;	/* useful for timing raw capture speed */
	    break;
	case 'E':
	    error_out = TRUE;
	    break;
	case '?':
	default:
	    usage(argv[0], NULL);
	    break;
	}

    /* expect no args left over */
    if (argv[optind]) {
	usage(argv[0], argv[optind]);
    }

    state = xil_open();
    if (state == NULL) {
	fprintf(stderr, "unable to open xil library\n");
	exit(1);
    }

    if (error_out)
	xil_install_error_handler(state, error_handler);

#ifdef PRE494
    if (!(rtvc_image = xil_create_from_device(state, "SUNWrtvc",
					      (void *) devname))) {
	fprintf(stderr, "failed to open SUNWrtvc device\n");
	xil_close(state);
	exit(1);
    }

    xil_set_device_attribute(rtvc_image, "PORT_V", (void *) port);
    if (image_skip)
	xil_set_device_attribute(rtvc_image, "IMAGE_SKIP", (void *) image_skip);
    if (max_buffers >= 0)
	xil_set_device_attribute(rtvc_image, "MAX_BUFFERS", (void *) max_buffers);

#else
    {
	XilDevice device;

	if (!(device = xil_device_create(state, "SUNWrtvc"))) {
	    fprintf(stderr, "Unable to create a device object\n");
	    xil_close(state);
	    exit(1);
	}

	xil_device_set_value(device, "DEVICE_NAME", (void *) devname);
	xil_device_set_value(device, "PORT_V", (void *) port);
	if (image_skip)
	    xil_device_set_value(device, "IMAGE_SKIP", (void *) image_skip);
	if (max_buffers >= 0)
	    xil_device_set_value(device, "MAX_BUFFERS", (void *) max_buffers);

	if (!(rtvc_image = xil_create_from_device(state, "SUNWrtvc",
						  device))) {
	    fprintf(stderr, "failed to open SUNWrtvc device\n");
	    xil_close(state);
	    exit(1);
	}
	xil_device_destroy(device);
    }
#endif

    {
	int format;

	xil_get_device_attribute(rtvc_image, "FORMAT_V", (void **) &format);
	if (format == 0) {
	    fprintf(stderr, "Unknown video format, exiting.\n");
	    xil_close(state);
	    exit(1);
	}
    }


    xil_get_info(rtvc_image, &width, &height, &nbands, &datatype);

    /* check for width or height too big */
    if (child_width > width || child_height > height) {
	fprintf(stderr, "Maximum width,height = %d,%d\n", width, height);
	xil_close(state);
	exit(1);
    }


    if ((child_width) || (child_height)) {
	if (!child_width)
	    child_width = width;
	if (!child_height)
	    child_height = height;
	if (!(rtvc_image = xil_create_child(rtvc_image,
					    (width - child_width) / 2,
					    (height - child_height) / 2,
					    child_width, child_height,
					    0, nbands))) {
	    fprintf(stderr, "failed to open child image on SUNWrtvc\n");
	    xil_close(state);
	    exit(1);
	}

	xil_get_info(rtvc_image, &width, &height, &nbands, &datatype);
    }

    if (grayscale) {
	rtvc_luma = xil_create_child(rtvc_image, 0, 0, width, height, 0, 1);
    }
    else {
	rescaled_image = xil_create(state, width, height, nbands, XIL_BYTE);
    }

    if (doscale) {
	scale_factor = 1.0 / doscale;
	width = (int) (((float) width * scale_factor) + 0.5);
	height = (int) (((float) height * scale_factor) + 0.5);
	if (!grayscale) {
	    scaled_image = xil_create(state, width, height, nbands, XIL_BYTE);
	}
    }
    else {
	scale_factor = 1.0;
    }
    if (grayscale) {
	/* create a 1-band image with the scaled dimensions */
	single_band = xil_create(state, width, height, 1, XIL_BYTE);
    }

    if (nodisplay) {
	/* for simplicity's sake, the 'nodisplay' option assumes a 1-banded */
	/* XIL_BYTE destination image */
	display = xil_create(state, width, height, 1, XIL_BYTE);
	display_depth = 8;
    }
    else {
	XSizeHints hints;

	int port;

	char *format;

	char titlebar[1024];

	/* xlib window creation */
	xdisplay = XOpenDisplay(NULL);
	if (!xdisplay) {
	    fprintf(stderr, "Unable to connect to server\n");
	    xil_close(state);
	    exit(1);
	}
	display_depth = DefaultDepth(xdisplay, DefaultScreen(xdisplay));
	xwindow = XCreateSimpleWindow(xdisplay, DefaultRootWindow(xdisplay),
				      winx, winy, width, height, 0, 0, 0);
	if (!xwindow) {
	    fprintf(stderr, "Unable to create window\n");
	    xil_close(state);
	    exit(1);
	}
	if (xyflag) {
	    hints.flags = USPosition;
	    hints.x = winx;
	    hints.y = winy;
	}
	else {
	    hints.flags = 0;
	}
	xil_get_device_attribute(rtvc_image, "FORMAT", (void **) &format);
	xil_get_device_attribute(rtvc_image, "PORT_V", (void **) &port);
	sprintf(titlebar, "%s: Port %d [%s]",
		(devname ? devname : "/dev/rtvc0"), port, format);

	XSetStandardProperties(xdisplay, xwindow, titlebar, "rtvc_display", 0,
			       argv, argc, &hints);

	/* if the user tries to shut down the window using the menu */
	if (closedownAtom = XInternAtom(xdisplay, "WM_DELETE_WINDOW", False))
	    XSetWMProtocols(xdisplay, xwindow, &closedownAtom, 1);

	/* list window events used */
	XSelectInput(xdisplay, xwindow, ExposureMask | ButtonPressMask);

	/* make the window visible */
	XMapWindow(xdisplay, xwindow);

	/* wait for the window to be mapped (an Expose event) */
	do
	    XNextEvent(xdisplay, &event);
	while (event.xany.type != Expose);

	display = xil_create_from_window(state, xdisplay, xwindow);
    }

    if (display_depth == 8) {
	if (grayscale) {
	    if (!nodisplay) {
		XilLookup grayramp;

		int num_entries = 256;

		int i;

		Xil_unsigned8 *graydata = malloc(3 * num_entries);

		for (i = 0; i < num_entries; i++) {
		    graydata[i * 3 + 2] = graydata[i * 3 + 1] = graydata[i * 3] = i;
		}
		grayramp = xil_lookup_create(state, XIL_BYTE, XIL_BYTE,
					     3, num_entries, 0,
					     graydata);
		init_cmap(grayramp, xdisplay, xwindow, 0);
	    }
	}
	else {
	    get_dither_info(state, rescale_values, offset_values, &mask,
			    &yuv_to_rgb, &colorcube);
	    if (!nodisplay) {
		init_cmap(yuv_to_rgb, xdisplay, xwindow, -1);
	    }
	}
    }
    else if (display_depth == 24) {
	if (grayscale) {
	    int i, t;

	    Xil_unsigned8 *data = malloc(256 * 3);

	    if (data == NULL) {
		fprintf(stderr, "Out of memory\n");
		xil_close(state);
		exit(1);
	    }
	    for (i = 0, t = 0; i < 256; i++, t += 3) {
		data[t] = i;
		data[t + 1] = i;
		data[t + 2] = i;
	    }
	    graylookup = xil_lookup_create(state, XIL_BYTE, XIL_BYTE,
					   3, 256, 0, data);
	}
	else {
	    cspace_ycc601 = xil_colorspace_get_by_name(state, "ycc601");
	    cspace_rgb709 = xil_colorspace_get_by_name(state, "rgb709");
	    if (doscale)
		xil_set_colorspace(scaled_image, cspace_ycc601);
	    else
		xil_set_colorspace(rtvc_image, cspace_ycc601);
	    xil_set_colorspace(display, cspace_rgb709);
	}
    }
    else {
	fprintf(stderr, "visual depth of %d is not supported\n", display_depth);
	xil_close(state);
	exit(1);
    }

    for (frames = 0; frames < total_frames; frames++) {
	if (grayscale) {
	    if (display_depth == 24) {

		/*
		 * even if scale_factor is 1, this combination of xil_scale()
		 * & xil_lookup() is faster than
		 * xil_lookup(rtvc_luma,display,graylookup)
		 */
		xil_scale(rtvc_luma, single_band, "nearest",
			  scale_factor, scale_factor);
		xil_lookup(single_band, display, graylookup);
		xil_toss(single_band);
	    }
	    else {		/* display_depth is 8 */

		/*
		 * even if scale_factor is 1, xil_scale() is faster than
		 * xil_copy(rtvc_luma,display) for this case
		 */
		xil_scale(rtvc_luma, display, "nearest",
			  scale_factor, scale_factor);
		if (nodisplay)
		    xil_sync(display);
	    }
	}
	else {
	    if (display_depth == 24) {
		if (doscale) {
		    xil_scale(rtvc_image, scaled_image, "nearest",
			      scale_factor, scale_factor);
		    xil_color_convert(scaled_image, display);
		    xil_toss(scaled_image);
		}
		else {
		    xil_color_convert(rtvc_image, display);
		}
	    }
	    else {		/* display_depth is 8 */
		xil_rescale(rtvc_image, rescaled_image, rescale_values,
			    offset_values);
		if (doscale) {
		    xil_scale(rescaled_image, scaled_image, "nearest",
			      scale_factor, scale_factor);
		    xil_ordered_dither(scaled_image, display, colorcube, mask);
		    xil_toss(scaled_image);
		}
		else {
		    xil_ordered_dither(rescaled_image, display,
				       colorcube, mask);
		}
		if (nodisplay)
		    xil_sync(display);
		xil_toss(rescaled_image);
	    }
	}
	/* don't include overhead of first frame in frame rate calculation */

#ifdef PRE494
	if (frames == 0)
	    gettimeofday(&t0);
#else
	if (frames == 0)
	    gettimeofday(&t0, NULL);
#endif

	if (!nodisplay) {
	    while (XPending(xdisplay)) {
		XNextEvent(xdisplay, &event);

		if (event.xany.type == ClientMessage &&
			((XClientMessageEvent *) & event)->data.l[0] == closedownAtom) {
		    goto cleanup;

		}
		else if (event.xany.type == ButtonPress) {
		    goto done;

		}
		else if (event.xany.type != Expose) {
		    fprintf(stderr, "event %d\n", event.xany.type);
		}
	    }
	}
    }

done:

#ifdef PRE494
    gettimeofday(&t1);
#else
    gettimeofday(&t1, NULL);
#endif

    t = (t1.tv_sec - t0.tv_sec) + (t1.tv_usec - t0.tv_usec) * 1e-6;
    printf("%d frames in %g seconds, %g frames/second\n",
	   frames - 1, t, (frames - 1) / t);

cleanup:

    if (!nodisplay) {
	/* must destroy display image before calling XCloseDisplay */
	xil_destroy(display);

	XCloseDisplay(xdisplay);
    }


    xil_close(state);

    return 0;
}

/* - - - - - - - - - - - - - - - - - - - - - - */
